# -----------------------------------------------------------------
# 
#   MELODY HARMONIZATION WITH INTERPOLATED PROBABILISTIC MODELS
# 
# This package contains the code for automatic melody harmonization
# described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Melody harmonisation
# with interpolated probabilistic models,” INRIA Technical
# Report RR-8110, 2012,
# URL: http://hal.inria.fr/hal-00742957/PDF/RR-8110.pdf
# 
# This code has been written by:
#  - Stanisław Raczyński, INRIA
#  - Satoru Fukayama, The University of Tokyo
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# Contact information: stanislaw.raczynski@inria.fr
# 
# -----------------------------------------------------------------

loadData = function(dataset="testing", datasetSuffix="_1_1_ignore")
{
  load(paste("data/dataset_", dataset, datasetSuffix, ".Rdata", sep=""))
  load(paste("data/models", datasetSuffix, ".Rdata", sep=""))
  return(list(allRootInts=allRootInts, allTypeInts=allTypeInts,
              allRootPrevInts=allRootPrevInts, allTypePrevInts=allTypePrevInts,
              allKeysInts=allKeysInts, templateSequence=templateSequence,
              chordDistribution=chordDistribution, tonalityModel=tonalityModel,
              melodyModel=melodyModel, chordModel=chordModel))
}

createUniformProbabilityMatrix = function(dim, dimnames)
{
  # This is a uniform probability matrix that does not allow mixing
  # unknown ("none") and known chord information
  uniformProbability = array(1, dim=dim, dimnames=dimnames)
  uniformProbability["none",] = 0
  uniformProbability[,"none"] = 0
  uniformProbability["none","none"] = 1
  uniformProbability = uniformProbability / sum(uniformProbability)

  return(uniformProbability)
}

calculateCrossEntropies = function(interpolationCoefficients,
                                   allData,
                                   interpolation="linear",  # {"linear", "log-linear"}
                                   parallelize=TRUE,  #TODO
                                   silent=FALSE,
                                   avoidInfinity=FALSE # useful when training with L-BFGS-B; adds 1e-16 before taking logs
                                  )
{
  # Negative coefficients appear due to numerical errors in L-BFGS-B, e.g.:
  # | Fudge factors = 0 0 2.945738 -1.110223e-16 0
  # | · Average likelihood = NaN
  # Other methods have no means of using the lower bound, so the following
  # is useful in both cases:
  interpolationCoefficients[interpolationCoefficients < 0] = 0

  if(!silent)
  {
    cat(" +", paste(rep("-", 70, collapse="")), "\n", sep="")
    cat(" | Fudge factors =", interpolationCoefficients, "\n")
  }

  if(avoidInfinity) {
    logDisturbance = 0 # .Machine$double.xmin    # 1e-24
  } else
    logDisturbance = 0
  dataLength = length(allData$allRootInts)

  uniformProbability = createUniformProbabilityMatrix(dim=dim(allData$chordDistribution),
                                 dimnames=dimnames(allData$chordDistribution))

  ##MOCK: to test if it would help if state tying in the prior would help
  #allData$chordDistribution = matrix(rep(colSums(allData$chordDistribution), each=13), nrow=13) / 13

  if(interpolation == "linear")
  {
    interpolationCoefficients = interpolationCoefficients / sum(interpolationCoefficients) # make sure they are normalized

    tonalityProbs = interpolationCoefficients[1] * allData$tonalityModel[cbind(allData$allRootInts, allData$allTypeInts, allData$allKeysInts)]
    melodyProbs   = interpolationCoefficients[2] * allData$melodyModel[cbind(allData$allRootInts, allData$allTypeInts, allData$templateSequence)]
    bigramProbs   = interpolationCoefficients[3] * allData$chordModel[cbind(allData$allRootInts, allData$allTypeInts, allData$allRootPrevInts, allData$allTypePrevInts)]
    unigramProbs  = interpolationCoefficients[4] * allData$chordDistribution[cbind(allData$allRootInts, allData$allTypeInts)]
    nogramProbs   = interpolationCoefficients[5] * rep(1/prod(dim(allData$chordDistribution)), dataLength)

    # Average log-probability (cross-entropy)
    probSum = tonalityProbs + melodyProbs + bigramProbs + unigramProbs + nogramProbs + logDisturbance
    modellikelihood = -sum(log2(probSum[probSum != 0])) / length(which(probSum != 0))
    if(modellikelihood == 0)
      modellikelihood = 1e24

  } else if(interpolation == "loglinear")
  {
    # This is a special special case with all coefficients equal to 0
    if(all(interpolationCoefficients[c(1,4,7)] == 0))
    {
      cat(" | · Average likelihood (special) =", 1e24, "\n")  # -log2(logDisturbance)
      return(1e24)
    }

    execTimes = rep(NA, 4)
    
    execTimes[1] = system.time({
      if(interpolationCoefficients[7] != 0) {
        bigramProbs = array(NA, dim=c(dim(allData$chordDistribution), dataLength))
        for(ii in seq_len(dataLength))
          bigramProbs[,,ii] = allData$chordModel[,,allData$allRootPrevInts[ii], allData$allTypePrevInts[ii]]
      }
    })[3]
    #    execTimes[1] = system.time({bigramProbs = sapply(seq_len(dataLength), function(ii) allData$chordModel[,,allData$allRootPrevInts[ii], allData$allTypePrevInts[ii]])})[3]
    
    # Apply smoothing (with normalized sets of coefficients)
    execTimes[2] = system.time({
      if(interpolationCoefficients[1] != 0) {
        tonalityProbs = interpolationCoefficients[1] * log((allData$tonalityModel[,,allData$allKeysInts] +
        interpolationCoefficients[2] * as.vector(allData$chordDistribution) +
        interpolationCoefficients[3] * as.vector(uniformProbability)) / sum(1, interpolationCoefficients[2:3]))
      } else {
        tonalityProbs = 0
      }
      
      if(interpolationCoefficients[4] != 0) {
        melodyProbs   = interpolationCoefficients[4] * log((allData$melodyModel[,,allData$templateSequence] +
        interpolationCoefficients[5] * as.vector(allData$chordDistribution) +
        interpolationCoefficients[6] * as.vector(uniformProbability)) / sum(1, interpolationCoefficients[5:6]))
      } else {
        melodyProbs = 0
      }
      
      if(interpolationCoefficients[7] != 0) {
        bigramProbs   = interpolationCoefficients[7] * log((bigramProbs +
        interpolationCoefficients[8] * as.vector(allData$chordDistribution) +
        interpolationCoefficients[9] * as.vector(uniformProbability)) / sum(1, interpolationCoefficients[8:9]))
      } else {
        bigramProbs = 0
      }
    })[3]
   
    # Log-linear interpolation
    execTimes[3] = system.time({
      probabilityArray = exp(tonalityProbs + melodyProbs + bigramProbs)
    })[3]
    
    # Normalization
    execTimes[4] = system.time({
      probabilityArray = probabilityArray / rep(apply(probabilityArray, 3, sum), each=prod(dim(probabilityArray)[1:2]))
      probabilityArray[is.nan(probabilityArray)] = 0  # NaNs appear if most of the coefficients are equal to 0,
                               # during log-linear interpolation: exp(log(0)*a) = NaN,
                               # or during normalization, if all probabilities for a given time are 0.
                               # Because of the above, when optimizing, use avoidInfinity=TRUE so
                               # that the optimization doesn't crash when the probability is 0 (likelihood is infinite).
    })[3]
    
    # Average log-probability (cross-entropy)
    probSum = probabilityArray[cbind(allData$allRootInts, allData$allTypeInts, seq_len(dataLength))]
    modellikelihood = -sum(log2(probSum[probSum != 0])) / length(which(probSum != 0))
    if(modellikelihood == 0)
      modellikelihood = 1e24

  }  else
  {
    stop("Unknown interpolation method.")
  }

  if(!silent) cat(" | · Average likelihood =", modellikelihood, "\n")
  return(modellikelihood)
}

createPlacementFunction = function(interpolation, tonalityModel, melodyModel, bigramModel)
{
  modelsUsed = c(tonalityModel, melodyModel, bigramModel)
  if(interpolation == "loglinear")
  {
    parameterMask = rep(modelsUsed*1, each=3)
    totalParameterCount = 9   # number of parameters used by 'calculateCrossEntropies'
  } else if(interpolation == "linear") {
    parameterMask = c(modelsUsed*1, 1, 1)
    totalParameterCount = 5
  } else
    stop("Unknown interpolation method.")

  # This functions puts the limited parameter set in their places in the
  # parameter vector as expected by the 'calculateCrossEntropies' function
  return(function(x) sapply(1:totalParameterCount, function(ii) if(parameterMask[ii] == 1) x[sum(parameterMask[1:ii])] else 0))
}
