# -----------------------------------------------------------------
# 
#   MELODY HARMONIZATION WITH INTERPOLATED PROBABILISTIC MODELS
# 
# This package contains the code for automatic melody harmonization
# described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Melody harmonisation
# with interpolated probabilistic models,” INRIA Technical
# Report RR-8110, 2012,
# URL: http://hal.inria.fr/hal-00742957/PDF/RR-8110.pdf
# 
# This code has been written by:
#  - Stanisław Raczyński, INRIA
#  - Satoru Fukayama, The University of Tokyo
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# Contact information: stanislaw.raczynski@inria.fr
# 
# -----------------------------------------------------------------

library("multicore")

source("harmonizeMelody.R")
source("settings.R")
source("datasetDefinitions.R")

source("prepareDataForTesting.R")

root_score = function(estimatedChords, groundTruth)
  1 * (estimatedChords[1,] == groundTruth[1,])

root_score_weighted = function(estimatedChords, groundTruth)
{
  diffs = abs(estimatedChords[1,] - groundTruth[1,])
            #  0  1  2  3  4  5  6  7  8  9  10 11
            #  0  11 10 9  8  7  6  5  4  3  2  1
  weights = c(10, 0, 1, 3, 3, 5, 0, 5, 3, 3, 1, 0, 0) / 10
  score = weights[diffs + 1]

  # Special case: subtract the misnamed non-chords (not real intervals there)
  score[(estimatedChords[1,] == 13) & (groundTruth[1,] != 13)] = 0
  score[(estimatedChords[1,] != 13) & (groundTruth[1,] == 13)] = 0

  return(score)
}

evaluate_roots = function(estimatedChords, groundTruth)
  sum(root_score(estimatedChords, groundTruth)) / ncol(groundTruth)

evaluate_roots_weighted = function(estimatedChords, groundTruth)
  sum(root_score_weighted(estimatedChords, groundTruth)) / ncol(groundTruth)

get_dyads_raw = function(estimatedChords, groundTruth)
{
  majorCompatibles = c("augmented", "augmented-ninth", "augmented-seventh",
                       "dominant", "dominant-11th", "dominant-13th",
                       "dominant-ninth", "dominant-seventh", "major",
                       "major-minor", "major-ninth", "major-seventh",
                       "major-sixth", "major-13th", "power", "suspended_second",
                       "suspended-fourth")
  minorCompatibles = c("diminished", "diminished-seventh", "half-diminished",
                       "minor-major", "minor", "minor-11th", "minor-ninth",
                       "minor-seventh", "minor-sixth")

  simplifyChords = function(chordTypeSequence)
  {
    chordTypeSequence[which(allPossibleChordTypes[chordTypeSequence] %in% majorCompatibles)] = which(allPossibleChordTypes == "major")[1]
    chordTypeSequence[which(allPossibleChordTypes[chordTypeSequence] %in% minorCompatibles)] = which(allPossibleChordTypes == "minor")[1]
    return(chordTypeSequence)
  }

  estimatedChordTypes = simplifyChords(estimatedChords[2,])
  groundTruthChordTypes = simplifyChords(groundTruth[2,])

  return(1 * (estimatedChordTypes == groundTruthChordTypes))
}

evaluate_dyads_raw = function(estimatedChords, groundTruth)
  sum(get_dyads_raw(estimatedChords, groundTruth)) / ncol(groundTruth)

evaluate_dyads = function(estimatedChords, groundTruth)
  sum(get_dyads_raw(estimatedChords, groundTruth) * evaluate_roots(estimatedChords, groundTruth)) / ncol(groundTruth)

evaluate_dyads_weighted = function(estimatedChords, groundTruth)
  sum(get_dyads_raw(estimatedChords, groundTruth) * evaluate_roots_weighted(estimatedChords, groundTruth)) / ncol(groundTruth)

# All possible key types:
#  [1]  "augmented"          "augmented-ninth"    "augmented-seventh"
#  [4]  "diminished"         "diminished-seventh" "dominant"
#  [7]  "dominant-11th"      "dominant-13th"      "dominant-ninth"
#  [10] "dominant-seventh"   "half-diminished"    "major"
#  [13] "major-minor"        "major-ninth"        "major-seventh"
#  [16] "major-sixth"        "minor"              "minor-11th"
#  [19] "minor-major"        "minor-ninth"        "minor-seventh"
#  [22] "minor-sixth"        "suspended-fourth"   "suspended-second"
#  [25] "none"               "major-13th"         "power"
get_triads_raw = function(estimatedChords, groundTruth)
{
  majorCompatibles = c("dominant", "dominant-11th", "dominant-13th",
                       "dominant-ninth", "dominant-seventh", "major",
                       "major-minor", "major-ninth", "major-seventh",
                       "major-sixth", "major-13th", "power")
  augmentedCompatibles = c("augmented", "augmented-ninth", "augmented-seventh")
  minorCompatibles = c("minor-major", "minor", "minor-11th", "minor-ninth",
                       "minor-seventh", "minor-sixth")
  diminishedCompatibles = c("diminished", "diminished-seventh", "half-diminished")
  sus2Compatibles = c("suspended-second")
  sus4Compatibles = c("suspended-fourth")

  simplifyChords = function(chordTypeSequence)
  {
    chordTypeSequence[which(allPossibleChordTypes[chordTypeSequence] %in% majorCompatibles)] = which(allPossibleChordTypes == "major")[1]
    chordTypeSequence[which(allPossibleChordTypes[chordTypeSequence] %in% minorCompatibles)] = which(allPossibleChordTypes == "minor")[1]
    chordTypeSequence[which(allPossibleChordTypes[chordTypeSequence] %in% augmentedCompatibles)] = which(allPossibleChordTypes == "augmented")[1]
    chordTypeSequence[which(allPossibleChordTypes[chordTypeSequence] %in% diminishedCompatibles)] = which(allPossibleChordTypes == "diminished")[1]
    chordTypeSequence[which(allPossibleChordTypes[chordTypeSequence] %in% sus2Compatibles)] = which(allPossibleChordTypes == "suspended-second")[1]
    chordTypeSequence[which(allPossibleChordTypes[chordTypeSequence] %in% sus4Compatibles)] = which(allPossibleChordTypes == "suspended-fourth")[1]
    return(chordTypeSequence)
  }

  estimatedChordTypes = simplifyChords(estimatedChords[2,])
  groundTruthChordTypes = simplifyChords(groundTruth[2,])

  return(1 * (estimatedChordTypes == groundTruthChordTypes))
}

evaluate_triads_raw = function(estimatedChords, groundTruth)
  sum(get_triads_raw(estimatedChords, groundTruth)) / ncol(groundTruth)

evaluate_triads = function(estimatedChords, groundTruth)
  sum(get_triads_raw(estimatedChords, groundTruth) * evaluate_roots(estimatedChords, groundTruth)) / ncol(groundTruth)

evaluate_triads_weighted = function(estimatedChords, groundTruth)
  sum(get_triads_raw(estimatedChords, groundTruth) * evaluate_roots_weighted(estimatedChords, groundTruth)) / ncol(groundTruth)

loadData = function(dataset, unitType=1, unitDivision=1)
{
  ii = 1
  allFiles = list()
  for(filename in dataset)
  {
    allFiles[[ii]] = prepareFile(filename, unitType=unitType, unitDivision=unitDivision)
    ii = ii + 1
  }
  return(allFiles)
}

getResults = function(allHarmonizations, reference, unitMultiplier=1)
simplify2array(lapply(1:length(allHarmonizations), function(ii) {
  estimatedChords = rbind(rep(allHarmonizations[[ii]][1,], each=unitMultiplier),
                          rep(allHarmonizations[[ii]][2,], each=unitMultiplier))
  groundTruth     = rbind(reference[[ii]][[3]], reference[[ii]][[4]])
  remainder = ncol(groundTruth) - ncol(estimatedChords)
  if(remainder < 0)
    stop("Negative remainder!")
  cat("R", remainder, " ", sep="")
  groundTruth = groundTruth[,1:ncol(estimatedChords)]
  return(c(evaluate_roots(estimatedChords, groundTruth),
           evaluate_roots_weighted(estimatedChords, groundTruth),
           evaluate_dyads_raw(estimatedChords, groundTruth),
           evaluate_dyads(estimatedChords, groundTruth),
           evaluate_dyads_weighted(estimatedChords, groundTruth),
           evaluate_triads_raw(estimatedChords, groundTruth),
           evaluate_triads(estimatedChords, groundTruth),
           evaluate_triads_weighted(estimatedChords, groundTruth)))
}))

getHarmonizations = function(allFiles, reference, interpolation, interpolationCoefficients, multicore=TRUE, unitDivision, unitMultiplier=1)
{
  allHarmonizations = c(lapply, mclapply)[[multicore+1]](1:length(allFiles), function(ii) {
    harmonizationSequence = harmonizeMelody(allFiles[[ii]][[1]], allFiles[[ii]][[2]], interpolation=interpolation, interpolationCoefficients=interpolationCoefficients, unitDivision=unitDivision)
    cat(ii, " ")
    return(harmonizationSequence)
  })
  cat("\n")

  allResults = getResults(allHarmonizations, reference ,unitMultiplier=unitMultiplier)

  return(list(allResults=allResults, allHarmonizations=allHarmonizations))
}
