# -----------------------------------------------------------------
# 
#   MELODY HARMONIZATION WITH INTERPOLATED PROBABILISTIC MODELS
# 
# This package contains the code for automatic melody harmonization
# described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Melody harmonisation
# with interpolated probabilistic models,” INRIA Technical
# Report RR-8110, 2012,
# URL: http://hal.inria.fr/hal-00742957/PDF/RR-8110.pdf
# 
# This code has been written by:
#  - Stanisław Raczyński, INRIA
#  - Satoru Fukayama, The University of Tokyo
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# Contact information: stanislaw.raczynski@inria.fr
# 
# -----------------------------------------------------------------

source("settings.R")
source("musicology.R")
source("calculateCrossEntropy.R")

maxInMatrix = function(a)
{
  maxPos = which.max(a)-1
  return(c( nrow=maxPos%%nrow(a)+1, ncol=floor(maxPos/nrow(a))+1))
}

posInMatrix = function(pos, nrow, ncol)
  c( nrow=(pos-1)%%nrow+1, ncol=floor((pos-1)/nrow)+1 )

harmonizeMelody = function(melody, keys, interpolation="loglinear", interpolationCoefficients=NULL, unitType=1, unitDivision=1)
{
  if(length(melody) != length(keys))
    stop("Melody and keys have different lengths")

  # Load the distributions and interpolation coefficients
  suffix = paste(unitType, "_", unitDivision, "_", stateTyingMethod, sep="")
  load(paste("data/models_", suffix, ".Rdata", sep=""))
  if(is.null(interpolationCoefficients))
  {
    load(paste("data/fudgeFactors_", suffix, ".Rdata", sep=""))
    if(interpolation == "linear") {
      interpolationCoefficients = fudgeFactors["par",1][[1]]
    } else
      interpolationCoefficients = fudgeFactors["par",6][[1]]
  }

  if(interpolation == "linear")
    interpolationCoefficients = interpolationCoefficients / sum(interpolationCoefficients)


  uniformProbability = createUniformProbabilityMatrix(dim=dim(chordDistribution),
                                 dimnames=dimnames(chordDistribution))

  rootCount   = dim(tonalityModel)[1]
  kindCount   = dim(tonalityModel)[2]
  keyCount    = dim(tonalityModel)[3]
  melodyCount = dim(melodyModel)[3]
  dataLength = length(melody)

  # Returns P(C1 | M1, K1)
  initialHarmonyDistribution = function(melody1, key1, interpolation, interpolationCoefficients)
  {
    if(interpolation == "linear")
    {
      # Make sure the coefficients are normalized
#      interpolationCoefficients = interpolationCoefficients / sum(interpolationCoefficients)
      
      probabilityArray = interpolationCoefficients[1] * tonalityModel[,,key1] +
                         interpolationCoefficients[2] * melodyModel[,,melody1] +
                         interpolationCoefficients[3] * chordModel[,,"none","none"] +
                         interpolationCoefficients[4] * chordDistribution +
                         interpolationCoefficients[5] * uniformProbability
    } else {
      smoothedTonalityModel = interpolationCoefficients[1] * log((tonalityModel[,,key1] +
                              interpolationCoefficients[2] * as.vector(chordDistribution) +
                              interpolationCoefficients[3] * as.vector(uniformProbability)) / sum(1, interpolationCoefficients[2:3]))
      smoothedTonalityModel[is.na(smoothedTonalityModel)] = -Inf
      
      smoothedMelodyModel = interpolationCoefficients[4] * log((melodyModel[,,melody1] +
                            interpolationCoefficients[5] * chordDistribution +
                            interpolationCoefficients[6] * uniformProbability) / sum(1, interpolationCoefficients[5:6]))
      smoothedMelodyModel[is.na(smoothedMelodyModel)] = -Inf
      
      smoothedBigramModel = interpolationCoefficients[7] * log((chordModel[,,"none","none"] +
                            interpolationCoefficients[8] * chordDistribution +
                            interpolationCoefficients[9] * uniformProbability) / sum(1, interpolationCoefficients[8:9]))
      smoothedBigramModel[is.na(smoothedBigramModel)] = -Inf
      
      probabilityArray = exp(smoothedTonalityModel + smoothedMelodyModel + smoothedBigramModel)

      # Normalization
      probabilityArray = probabilityArray / sum(probabilityArray)
    }
    return(log(probabilityArray))
  }

  # Returns P(Ct | Ct-1, Mt, K1)
  harmonyDistribution = function(melodyCombination, keyNr, interpolation, interpolationCoefficients)
  {
    if(interpolation == "linear")
    {
      # Make sure the coefficients are normalized
#      interpolationCoefficients = interpolationCoefficients / sum(interpolationCoefficients)

      probabilityArray = array(NA, dim=c(rootCount, kindCount, rootCount, kindCount))

      nonconditionalModel = as.vector(interpolationCoefficients[1] * tonalityModel[,,keyNr] +
                                      interpolationCoefficients[2] * melodyModel[,,melodyCombination] +
                                      interpolationCoefficients[4] * chordDistribution +
                                      interpolationCoefficients[5] * uniformProbability)

      for(rootNr in 1:rootCount) for(kindNr in 1:kindCount)
        probabilityArray[,,rootNr,kindNr] = interpolationCoefficients[3] * chordModel[,,rootNr,kindNr] + nonconditionalModel

    } else {
      ##TODO: These should be precalculated!!!
      smoothedTonalityModel = interpolationCoefficients[1] * log((tonalityModel[,,keyNr] +
                              interpolationCoefficients[2] * as.vector(chordDistribution) +
                              interpolationCoefficients[3] * as.vector(uniformProbability)) / sum(1, interpolationCoefficients[2:3]))
      smoothedTonalityModel[is.na(smoothedTonalityModel)] = -Inf
      
      smoothedMelodyModel = interpolationCoefficients[4] * log((melodyModel[,,melodyCombination] +
                            interpolationCoefficients[5] * chordDistribution +
                            interpolationCoefficients[6] * uniformProbability) / sum(1, interpolationCoefficients[5:6]))
      smoothedMelodyModel[is.na(smoothedMelodyModel)] = -Inf
      
      smoothedBigramModel = interpolationCoefficients[7] * log((chordModel +
                            interpolationCoefficients[8] * as.vector(chordDistribution) +
                            interpolationCoefficients[9] * as.vector(uniformProbability)) / sum(1, interpolationCoefficients[8:9]))
      smoothedBigramModel[is.na(smoothedBigramModel)] = -Inf
      
      probabilityArray = array(NA, dim=c(rootCount, kindCount, rootCount, kindCount))
      for(rootNr in 1:rootCount) for(kindNr in 1:kindCount)
      {
        probabilityArray[,,rootNr,kindNr] = exp(smoothedTonalityModel +
                                                smoothedMelodyModel +
                                                smoothedBigramModel[,,rootNr,kindNr])
        probabilityArray[,,rootNr,kindNr] = probabilityArray[,,rootNr,kindNr] / sum(probabilityArray[,,rootNr,kindNr])  # normalization
      }
    }
    return(log(probabilityArray))
  }

  # Empty structures for later
  idx_C = array(NA, dim=c(rootCount, kindCount, dataLength))

  # ---- First frame ----
  # P(C1)
  prob0 = initialHarmonyDistribution(melody[1], keys[1], interpolation=interpolation, interpolationCoefficients)

  # ---- Middle & last frames ----
  if(dataLength > 1)
    for(timeNr in 2:dataLength)
    {
        # P(Ct, Ct-1)
        prob1 = harmonyDistribution(melody[timeNr], keys[timeNr], interpolation=interpolation, interpolationCoefficients)
        for(rootNr in 1:rootCount) for(kindNr in 1:kindCount)
          prob1[rootNr,kindNr,,] = prob1[rootNr,kindNr,,] + prob0

        # P(Ct)
        idx_C[,,timeNr-1] = apply(prob1, c(1,2), which.max)
        prob0             = apply(prob1, c(1,2), max)
  }

  # ---- Backtracing ----
  harmonySequence = matrix(NA, 2, dataLength)
  rownames(harmonySequence) = c("root", "kind")

  harmonySequence[, dataLength] = maxInMatrix(prob0)
  if(dataLength > 1)
    for(timeNr in (dataLength-1):1)
      harmonySequence[, timeNr] = posInMatrix(idx_C[harmonySequence[1, timeNr+1], harmonySequence[2, timeNr+1], timeNr] ,rootCount, kindCount)

  return(harmonySequence)
}

harmonizeFile = function(filename, interpolation="loglinear", interpolationCoefficients=NULL, unitType=1, unitDivision=1, ignoreOnsetMarks=TRUE)
{
  # Convert the data into a datastream
  #system(paste("python", "convertToDataStreams.py", paste("leadsheets", filename, sep="/"), "tempDataStream", unitType, unitDivision))
  system("sed -i \"s/(//g;s/)//g;s/'//g;s/,$//g\" tempDataStream.melody")
  system("sed -i \"s/(//g;s/)//g;s/'//g;s/ //g\" tempDataStream.tonality")

  currentData = strsplit(readLines("tempDataStream.melody"), ", ")
  if(ignoreOnsetMarks)
    currentData = lapply(currentData, function(noteNames) sub("[0-9]+", "", noteNames))

  # Creating the activity matrix
  activityMatrix = sapply(currentData, function(x) 1*(pitchClasses[-melodyPitchClassCount] %in% x))
  if(stateTyingMethod == "ignore")
  {
    activityMatrix = t(sapply(0:11, function(pitchClassNr) colSums(matrix(activityMatrix[pitchClassNrs[-length(pitchClassNrs)] == pitchClassNr,], ncol=ncol(activityMatrix)), na.rm=TRUE)))
    activityMatrix[activityMatrix > 1] = 1
  } else {
    rownames(activityMatrix) = pitchClasses[-length(pitchClasses)]
  }

  # Finding activity vectors in the dictionary
  maxConcurrentNotes = 5 # max(colSums(activityMatrix))
  load(paste("data/dictionaryMatrix_", stateTyingMethod, "_", maxConcurrentNotes, ".Rdata", sep=""))
  dictionarySizes = colSums(dictionaryMatrix)
  activitySizes = colSums(activityMatrix)
  templateSequence = rep(NA, ncol(activityMatrix))
  for(vectorNr in 1:ncol(activityMatrix))
  {
    if((vectorNr %% 10000) == 0) cat(".")
      for(templateNr in which(dictionarySizes == activitySizes[vectorNr]))
        if(all(dictionaryMatrix[,templateNr] == activityMatrix[,vectorNr]))
        {
          templateSequence[vectorNr] = templateNr
          break
        }
  }

  # Load the tonality sequence
  keyData = read.table("tempDataStream.tonality", sep=",", comment.char="", stringsAsFactors=FALSE)
  keyData = data.frame(tonic=factor(paste(c("C","D","E","F","G","A","B")[keyData[,1]+1], c("b","","#")[keyData[,2]+2], sep=""), levels=pitchClasses), mode=keyData[,3])
  keyData[,1] = factor(possiblePitchClasses[pitchClassNrs[keyData[,1]]+1], levels=possiblePitchClasses[1:12])
  possibleKeys = c(paste(possiblePitchClasses[1:12], "major", sep="-"), paste(possiblePitchClasses[1:12], "minor", sep="-"))
  keyData = factor(paste(keyData[,1], keyData[,2], sep="-"), levels=possibleKeys)
  keys = as.integer(keyData)

  harmonySequence = harmonizeMelody(templateSequence, keys, interpolation=interpolation, interpolationCoefficients=interpolationCoefficients, unitType=unitType, unitDivision=unitDivision)

  return(harmonySequence)
}

evaluateFile = function(filename, interpolation="loglinear", interpolationCoefficients=NULL, unitType=1, unitDivision=1)
{
  system(paste("python", "convertToDataStreams.py", paste("leadsheets", filename, sep="/"),  "tempDataStream", unitType, unitDivision))
  system("sed -i \"s/(//g;s/)//g;s/'//g;s/ //g\" tempDataStream.chord")

  # Load the chord sequence (for evaluation)
  chordData = read.table("tempDataStream.chord", sep=",", comment.char="", stringsAsFactors=FALSE)

  allRoots = chordData[,1]
  allRoots[is.na(allRoots)] = "none"
  allRoots = pitchClassNrs[allRoots]
  allRoots[is.na(allRoots)] = 12
  names(allRoots) = possiblePitchClasses[allRoots+1]
  allRootInts = as.integer(allRoots)+1

  allTypes = factor(chordData[,2], levels=allPossibleChordTypes)
  allTypes[is.na(allTypes)] = "none"
  allTypeInts = as.integer(allTypes)

  harmonySequence = harmonizeFile(filename=filename, interpolation=interpolation, interpolationCoefficients=interpolationCoefficients, unitType=unitType, unitDivision=unitDivision)

  unlink("tempDataStream.*")

  accuracy = length(which(harmonySequence[1,] == allRootInts)) / length(allRootInts)

  cat("Estimated sequence:  ", harmonySequence[1,], "\n")
  cat("Groundtruth sequence:", allRootInts, "\n")
  cat("Accuracy: ", round(100*accuracy, 1), "%\n", sep="")

  return(accuracy)
}

#testH = harmonizeFile("musicxml.mxl.1234.xml", "linear")
#testH2 = harmonizeFile("musicxml.mxl.1234.xml", "loglinear")
#testH3 = harmonizeFile("musicxml.mxl.1234.xml", "linear")

