# -----------------------------------------------------------------
# 
#   MELODY HARMONIZATION WITH INTERPOLATED PROBABILISTIC MODELS
# 
# This package contains the code for automatic melody harmonization
# described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Melody harmonisation
# with interpolated probabilistic models,” INRIA Technical
# Report RR-8110, 2012,
# URL: http://hal.inria.fr/hal-00742957/PDF/RR-8110.pdf
# 
# This code has been written by:
#  - Stanisław Raczyński, INRIA
#  - Satoru Fukayama, The University of Tokyo
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# Contact information: stanislaw.raczynski@inria.fr
# 
# -----------------------------------------------------------------

library("plotrix")

source("harmonizationEvaluation.R")

fileNrs = (1:100)[-c(6,8,41)]   # files with incomplete annotations (bad references)

mysd = function(x, lengths) sqrt(length(lengths) * sum(((x - mean(x))^2) * lengths) / ((length(lengths)-1) * sum(lengths)))

confInt = function(results, confLevel=0.95)
  qnorm(1 - (1-confLevel)/2)*mysd(results, fileLengths)/sqrt(length(results))

# Reference
unitType = 1
unitDivision = 1
allFiles = loadData(leadsheets.testing, unitDivision=unitDivision)
fileLengths = sapply(allFiles, function(x) length(x[[1]]))[fileNrs]

conversionTable        =
c(10, 5,  3,  1,  8,  6,  11,  12, 9,   4,   2,   9,   7,   5,   7,   12,  2,
  3,    10,   8,    4)
names(conversionTable) =
c("A","E","D","C","G","F","Bb","B","G#","Eb","Db","Ab","Gb","Fb","F#","Cb","C#",
  "Ebb","Bbb","Abb","D#")
referenceResults = list()
unitMultiplier = 6/unitDivision
for(fileNr in 1:length(leadsheets.testing))
{
  referenceRoots = conversionTable[scan(paste("HA/harmonic-analysis/analyses/", leadsheets.testing[fileNr], ".chords", sep=""), what=character())]
  upBeat = unitMultiplier*length(allFiles[[fileNr]][[1]]) - length(referenceRoots)
  if(upBeat >= 0) {
    referenceRoots = c(rep(13, upBeat), referenceRoots)
  } else {
    warning("Negative upbeat!")
    referenceRoots = referenceRoots[(1-upBeat):length(referenceRoots)]
  }
  referenceRoots = apply(rbind(matrix(referenceRoots, nrow=unitMultiplier), -1), 2, median)
  referenceTypes = c(rep(25, length(which(referenceRoots == 13))),
                     rep(12, length(referenceRoots) - length(which(referenceRoots == 13))))
                     # 25 = none, 12 = major

  referenceResults[[fileNr]] = rbind(referenceRoots, referenceTypes)
}
allResults_ref = getResults(referenceResults, allFiles)
averageResults_ref = colSums(t(allResults_ref[,fileNrs]) * fileLengths) / sum(fileLengths)
confInt_ref = confInt(allResults_ref[1,fileNrs])

# Interpolated models
unitDivisions = c(1/16,1/8,1/4,1/2,1) # 1/32,
averageResults_lin = averageResults_loglin = matrix(NA, 8, length(unitDivisions))
confInts = matrix(NA, 2, length(unitDivisions))
reference = allFiles
for(ii in seq_along(unitDivisions))
{
  unitDivision = unitDivisions[ii]
#   if(ii == 1) {  # some files are too short for 1/32
#     allFiles = loadData(leadsheets.testing[(1:100)[-c(35,62,66,88)]], unitDivision=unitDivision)
#     fileNrs = (1:96)[-c(6,8,40)]
#   } else {
    allFiles = loadData(leadsheets.testing, unitDivision=unitDivision)
    fileNrs = (1:100)[-c(6,8,41)]
#   }
  unitMultiplier = 1/unitDivision

#  load(paste("data/fudgeFactors_", unitType, "_", unitDivision, "_ignore.Rdata", sep=""))
#  loglinearCoefficients = fudgeFactors["par",6][[1]]
#  linearCoefficients = fudgeFactors["par",1][[1]]
  if(unitDivision == 1) {             # 1 beat per frame
    loglinearCoefficients = c(1.090, 0.4, 0.05, 0.600, 0.1, 0.6, 0.012, 0.1, 0.0)
    linearCoefficients    = c(0.5,              0.1,              0.1,  0.1, 0.1)
  } else if(unitDivision == 1/2) {    # 2 beats per frame
    loglinearCoefficients = c(0.565, 0.4, 0.1,  0.440, 0.1, 0.5,  0.00825, 0, 0.02)
    linearCoefficients    = c(0.5,              0.18,             0.13,    0.1, 0.5)
  } else if(unitDivision == 1/4) {    # 4 beats per frame
    loglinearCoefficients = c(0.710, 0.3, 0.001, 0.100, 0.1, 0.5,  0.000001, 0, 0.02)
    linearCoefficients    = c(0.5,               0.01,             0.01,    0.1, 0)
  } else if(unitDivision == 1/8) {    # 8 beats per frame
    loglinearCoefficients = c(0.001, 0.0, 0.0,  0.000001, 0.01, 0, 0.000001, 0, 0.02)
    linearCoefficients    = c(0.1,              0.03,             0.01,    0.1, 0)
  } else if(unitDivision == 1/16) {   # 16 beats per frame
    loglinearCoefficients = c(0.05, 0.5, 0.1, 0.0001, 0, 0.1, 0.07, 0, 0.1)
    linearCoefficients    = c(0.5, 0.03, 0.12, 0, 0)
  } else if(unitDivision == 1/32) {   # 32 beats per frame
    loglinearCoefficients = c(0.01, 0.1, 0.1, 0.01, 0.1, 0.1, 0.03, 0, 0.1)
    linearCoefficients    = c(0.5, 0.0001, 0.0001, 0, 0)
  }
  loglin = getHarmonizations(allFiles, reference, "loglinear", loglinearCoefficients, unitDivision=unitDivision, unitMultiplier=unitMultiplier)
  lin    = getHarmonizations(allFiles, reference, "linear", linearCoefficients, unitDivision=unitDivision, unitMultiplier=unitMultiplier)

  fileLengths = sapply(allFiles, function(x) length(x[[1]]))[fileNrs] * unitMultiplier
  averageResults_lin[,ii]    = colSums(t(lin$allResults[,fileNrs]) * fileLengths) / sum(fileLengths)
  averageResults_loglin[,ii] = colSums(t(loglin$allResults[,fileNrs]) * fileLengths) / sum(fileLengths)

  confInts[,ii] = apply(cbind(lin$allResults[1,fileNrs], loglin$allResults[1,fileNrs]), 2, confInt)
}

source("Plots/plotAllHarmonyResults.R")
