# -----------------------------------------------------------------
# 
#   MELODY HARMONIZATION WITH INTERPOLATED PROBABILISTIC MODELS
# 
# This package contains the code for automatic melody harmonization
# described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Melody harmonisation
# with interpolated probabilistic models,” INRIA Technical
# Report RR-8110, 2012,
# URL: http://hal.inria.fr/hal-00742957/PDF/RR-8110.pdf
# 
# This code has been written by:
#  - Stanisław Raczyński, INRIA
#  - Satoru Fukayama, The University of Tokyo
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# Contact information: stanislaw.raczynski@inria.fr
# 
# -----------------------------------------------------------------

#########################################################
## Look-up tables
# Pitch classes found in the leadsheet data:
# training:   'A', 'A#', 'Ab', 'B', 'B#', 'Bb', 'C', 'C#',        'Cb', 'D', 'D#', 'Db', 'Dbb', 'E', 'E#', 'Eb', 'F', 'F#', 'Fb', 'G', 'G#', 'Gb'
# testing:    'A', 'A#', 'Ab', 'B', 'B#', 'Bb', 'C', 'C#', 'C##', 'Cb', 'D', 'D#', 'Db',        'E', 'E#', 'Eb', 'F', 'F#',       'G', 'G#', 'Gb'
# validation: 'A', 'A#', 'Ab', 'B', 'B#', 'Bb', 'C', 'C#',        'Cb', 'D', 'D#', 'Db',        'E', 'E#', 'Eb', 'F', 'F#',       'G', 'G#', 'Gb'
# We use them all:
# pitchClasses  = c('A', 'A#', 'Ab', 'B', 'B#', 'Bb', 'C', 'C#', 'C##', 'Cb', 'D', 'D#', 'Db', 'Dbb', 'E', 'E#', 'Eb', 'F', 'F#', 'Fb', 'G', 'G#', 'Gb', 'none')
# pitchClassNrs = c( 9,   10,   8,    11,  0,    10,   0,   1,    2,     11,   2,   3,    1,    0,     4,   5,    3,    5,   6,    4,    7,   8,    6, NA)
# names(pitchClassNrs) = pitchClasses

pitchClasses  = c('B#', 'C', 'Dbb', 'C#', 'Db', 'C##', 'D', 'D#', 'Eb', 'E', 'Fb', 'E#', 'F', 'F#', 'Gb', 'G', 'G#', 'Ab', 'A', 'A#', 'Bb', 'B', 'Cb', 'none')
pitchClassNrs = c(0,0,0,1,1,2,2,3,3,4,4,5,5,6,6,7,8,8,9,10,10,11,11,NA)
names(pitchClassNrs) = pitchClasses

accidentalTable = c("bbb", "bb", "b", "", "#", "##", "###")

possibleMajorKeys = c("Cb", "Gb", "Db", "Ab", "Eb", "Bb", "F", "C", "G", "D", "A", "E", "B", "F#", "C#")
possibleMinorKeys = c("Ab", "Eb", "Bb", "F", "C", "G", "D", "A", "E", "B", "F#", "C#", "G#", "D#", "A#")

allPossibleChordTypes = c("augmented", "augmented-ninth", "augmented-seventh",
   "diminished", "diminished-seventh", "dominant",
   "dominant-11th", "dominant-13th", "dominant-ninth",
   "dominant-seventh", "half-diminished", "major",
   "major-minor", "major-ninth", "major-seventh",
   "major-sixth", "minor", "minor-11th",
   "minor-major", "minor-ninth", "minor-seventh",
   "minor-sixth", "suspended-fourth", "suspended-second", "none",
  "major-13th", "power")

#########################################################
## Functions

flatsSharps = function(keyName, possibleKeys) which(possibleKeys == keyName) - 8

getAccidentals = function(pitchClass)
which(.accidentalTable == substr(pitchClass, 2, nchar(pitchClass))) - 4

hasAccidental = function(pitchClass, accidental) length(grep(accidental, pitchClass, fixed=TRUE)) > 0

notesInKey = function(keyName, mode)
{
  pitchClassList = c("C", "D", "E", "F", "G", "A", "B")
  keyStart = which(pitchClassList == substr(keyName, 1, 1))
  sharps = c(4, 1, 5, 2, 6, 3, 7)
  flats  = rev(sharps)
  if(mode == "major")
    positionInCOF = flatsSharps(keyName, possibleMajorKeys)
  else if(mode == "minor")
    positionInCOF = flatsSharps(keyName, possibleMinorKeys)
  else
    stop("Unknown key mode.")
  if(positionInCOF > 0)
    for(whereToAddSharp in sharps[(((1:positionInCOF)-1) %% 7) + 1])
      pitchClassList[whereToAddSharp] = paste(pitchClassList[whereToAddSharp], "#", sep="")
  if(positionInCOF < 0)
    for(whereToAddFlat in flats[(((1:-positionInCOF)-1) %% 7) + 1])
      pitchClassList[whereToAddFlat] = paste(pitchClassList[whereToAddFlat], "b", sep="")
  if(keyStart == 1)
    return(pitchClassList)
  else
    return(pitchClassList[c(keyStart:7, 1:(keyStart-1))])
}

# Returns the degree and accidentals of a pitch given a key.
#
# E.g. C is Ib under the key of C#-minor:
# 
# > getKeyDegree("C", "C#", "minor")
# $degree
# [1] 1
# 
# $accidentals
# [1] -1
#
getKeyDegree = function(pitchClass, tonic, mode)
{
  degree = which(substr(notesInKey(tonic, mode), 1, 1) == substr(pitchClass, 1, 1))
  return(list("degree" = degree,
              "accidentals" = getAccidentals(pitchClass) - getAccidentals(tonic)))
}

if(stateTyingMethod == "ignore") {
  possiblePitchClasses = c("C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B", "none")
  melodyPitchClassCount = 12
} else {
  melodyPitchClassCount = length(pitchClasses)-1 # exclude the 'none'
  possiblePitchClasses = pitchClasses
}
