# -----------------------------------------------------------------
# 
#   MELODY HARMONIZATION WITH INTERPOLATED PROBABILISTIC MODELS
# 
# This package contains the code for automatic melody harmonization
# described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Melody harmonisation
# with interpolated probabilistic models,” INRIA Technical
# Report RR-8110, 2012,
# URL: http://hal.inria.fr/hal-00742957/PDF/RR-8110.pdf
# 
# This code has been written by:
#  - Stanisław Raczyński, INRIA
#  - Satoru Fukayama, The University of Tokyo
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# Contact information: stanislaw.raczynski@inria.fr
# 
# -----------------------------------------------------------------

#######################################################

unitTypes     = c(0,1)
unitDivisions = c(1/16, 1/8, 1/4, 1/2, 1, 2, 3)
unitNames = c("16", "8", "4", "2", "1", "1/2", "1/3")

#######################################################

source("settings.R")
source("calculateCrossEntropy.R")
load("data/models_0_1_ignore.Rdata") # used in plotting

loadAllFactors = function(unitType, unitDivision, interpolation, modelNr=1)
{
  load(paste("data/fudgeFactors_", unitType, "_", unitDivision, "_", stateTyingMethod, ".Rdata", sep=""))
  return(fudgeFactors["par",][[c(linear=modelNr,loglinear=modelNr+5)[interpolation]]])
}

loadResults = function(unitType, unitDivision)
{
  load(paste("data/fudgeFactors_", unitType, "_", unitDivision, "_", stateTyingMethod, ".Rdata", sep=""))

  return(cbind(fudgeFactors["par",][[1]][1:3],
               fudgeFactors["par",][[6]][c(1,4,7)]))
}

modelCombinations = cbind(
  rep(c(1,1,0,0,1) == 1, 2),
  rep(c(1,0,1,0,1) == 1, 2),
  rep(c(1,0,0,1,0) == 1, 2)
)
modelNames = sapply(1:nrow(modelCombinations), function(ii) paste(paste(c("M","T","B")[modelCombinations[ii,]], collapse="")))[1:5]

getAllCE = function(modelNr)
  mapply(function(unitType, unitDivision, interpolation)
         {
           cat("1/", unitDivision, " of a ", c("quarter note", "beat")[unitType+1], ", interpolation = ", interpolation, ", model = ", modelNames[modelNr], "\n", sep="")
           placeParameters = createPlacementFunction(interpolation,
                                                     melodyModel=modelCombinations[modelNr,1],
                                                     tonalityModel=modelCombinations[modelNr,2],
                                                     bigramModel=modelCombinations[modelNr,3])
           fudgeFactors = loadAllFactors(unitType, unitDivision, interpolation=interpolation, modelNr=modelNr)
           cat(fudgeFactors, "\n")
           return(calculateCrossEntropies(placeParameters(fudgeFactors),
                                          loadData(dataset="testing", datasetSuffix=paste("_", unitType, "_", unitDivision, "_", stateTyingMethod, sep="")),
                                            interpolation=interpolation,
                                            avoidInfinity=TRUE))
         },
         unitType=rep(rep(unitTypes, each=length(unitDivisions)), 2),  # 2 interpolation methods
         unitDivision=rep(unitDivisions, length(unitTypes)*2),
         interpolation=rep(c("linear", "loglinear"), each=length(unitTypes)*length(unitDivisions))
      )

allCrossEntropies = sapply(1:5, getAllCE)
colnames(allCrossEntropies) = modelNames

# Cross-entropies of the prior only
load("data/priorOnlyEntropies.Rdata")
priorOnlyEntropies = priorOnlyEntropies[,2:8]

# Cross-entropy for the uniform distribution
perUnitLimit = -log2(createUniformProbabilityMatrix(dim=dim(chordDistribution), dimnames=dimnames(chordDistribution))[1])

##############################################################################
## Plotting

modelColors = c("red", "dark green", "blue", "black", "purple")

quarterData = cbind(allCrossEntropies[1:length(unitDivisions),], allCrossEntropies[(2*length(unitDivisions)+1):(3*length(unitDivisions)),])
beatData = cbind(allCrossEntropies[(length(unitDivisions)+1):(2*length(unitDivisions)),], allCrossEntropies[(3*length(unitDivisions)+1):(4*length(unitDivisions)),])
rownames(quarterData) = rownames(beatData) = unitNames

plotMarks = c(7,8,9,12,3)
lineTypes = c(1,rep(2,3),1)

plotModelLegend = function(x,y)
  legend(x, y, c(sapply(1:nrow(modelCombinations), function(ii) paste(c(paste(c("M","T","B")[modelCombinations[ii,]], collapse="+"))))[1:5], "uniform chord distribution", "prior only"), lty=c(lineTypes,2,5), col=c(modelColors, "black", "black"), bg="white", lwd=c(rep(1,5),2,1), pch=c(plotMarks,NA,NA), box.lwd=2, box.col="dark gray", text.width=1.5)

plotGridLines = function()
{
  abline(h=seq(0, 10, by=1), lty="dotted")
  abline(v=1/unitDivisions, lty="dotted")
}

plotModels = function(dataToPlot, ylim, main, ylab="Average cross-entropy [bits]")
{
  matplot(x=1/unitDivisions, y=dataToPlot, type="b", xlab="Frame length [beats]", ylab=ylab, main=main, pch=plotMarks, col=modelColors, ylim=ylim, axes=F, log="x", lty=lineTypes)
  axis(2)
  axis(1, at=1/unitDivisions, labels=rownames(beatData))
  box()
}

pdf("Plots/crossEntropy_perBeat.pdf", width=8, height=8)
{
  plotModels(beatData[,6:10] * unitDivisions, ylim=c(0,10), main="Average cross-entropy per beat for log-linear interpolation", ylab="Average cross-entropy [bits/beat]")
  plotGridLines()
  lines(1/unitDivisions, perUnitLimit * unitDivisions, pch=25, lty=4, lwd=2) # uniform
  lines(1/unitDivisions, priorOnlyEntropies[2,] * unitDivisions, lty=5, lwd=1) # prior
  plotModelLegend(2.3,9.5)
}
dev.off()

pdf("Plots/crossEntropy_perFrame.pdf", width=8, height=8)
{
  plotModels(beatData[,6:10], ylim=c(0,10), main="Average cross-entropy per frame for log-linear interpolation", ylab="Average cross-entropy [bits/frame]")
  plotGridLines()
  abline(h=perUnitLimit, lty=4, lwd=2) # uniform
  lines(1/unitDivisions, priorOnlyEntropies[2,], lty=5, lwd=1) # prior
  plotModelLegend(2.3,9.5)
}
dev.off()

pdf("Plots/crossEntropy_interpolation.pdf", width=8, height=8)
{
  plotModels(beatData[,1:5] - beatData[,6:10], ylim=c(-0.05,0.75), main="Per-frame entropy reduction of log-linear over linear interpolation", ylab="Average cross-entropy [bits/frame]")
  
  abline(h=seq(-0.0, 0.75, by=0.1), lty="dotted")
  abline(h=0, lty="solid")
  abline(v=1/unitDivisions, lty="dotted")
  
  legend(0.4, 0.58, sapply(1:nrow(modelCombinations), function(ii) paste(paste(c("M","T","B")[modelCombinations[ii,]], collapse="+")))[1:5], lty=1, col=modelColors, bg="white", pch=plotMarks)
}
dev.off()

pdf("Plots/crossEntropy_unit.pdf", width=8, height=8)
{
  matplot(x=1/unitDivisions, y=cbind(quarterData[,6],beatData[,6]), type="b", xlab="Log unit length [beats, quarter notes]", ylab="Average cross-entropy [bits]", main="Per-frame entropy for beat- and quarter note-based units", lty=c(2,1), pch=3, col="black", axes=F, log="x")
  axis(2); axis(1, at=1/unitDivisions, labels=rownames(beatData)); box()

  abline(h=seq(0, 3, by=0.5), lty="dotted")
  abline(v=1/unitDivisions, lty="dotted")

  legend(0.4, 1.5, c("quater notes", "beats"), lty=c(2,1), col="black", lwd=1, bg="white")
}
dev.off()
