# -----------------------------------------------------------------
# 
#   MELODY HARMONIZATION WITH INTERPOLATED PROBABILISTIC MODELS
# 
# This package contains the code for automatic melody harmonization
# described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Melody harmonisation
# with interpolated probabilistic models,” INRIA Technical
# Report RR-8110, 2012,
# URL: http://hal.inria.fr/hal-00742957/PDF/RR-8110.pdf
# 
# This code has been written by:
#  - Stanisław Raczyński, INRIA
#  - Satoru Fukayama, The University of Tokyo
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# Contact information: stanislaw.raczynski@inria.fr
# 
# -----------------------------------------------------------------

library("gtools")

source("settings.R")

source("datasetDefinitions.R")
getFileNumber = function(filename) sub("^[^0-9]+([0-9]+)[^0-9]+$", "\\1", filename)

source("musicology.R")

if(!ignoreOnsetMarks) {
  possiblePitchClasses = paste(pitchClasses, rep(c(0,1), each=melodyPitchClassCount))
  melodyPitchClassCount = 2 * melodyPitchClassCount
}

precalculateDictionaryMatrix = function(maxConcurrentNotes)
{
  outputFilename = paste("data/dictionaryMatrix_", stateTyingMethod, "_", maxConcurrentNotes, ".Rdata", sep="")
  
  if(file.exists(outputFilename))
  {
    cat("Dictionary matrix already exists. To re-calculate, delete file", outputFilename, "\n")
    return()
  }
  
  cat("Creating a dictionary of pitch class combinations...\n")
  howManyPitchTemplates = function(pitchClassCount, templateSize)
    round(factorial(pitchClassCount)/(factorial(templateSize)*factorial(pitchClassCount-templateSize)))
  howManyPitchTemplatesInTotal = function(pitchClassCount)
    sum(howManyPitchTemplates(pitchClassCount, 1:maxConcurrentNotes)) + 1

  dictionaryMatrix = matrix(0, melodyPitchClassCount, 1)
  for(templateSize in 1:maxConcurrentNotes)
  {
    templateCount = howManyPitchTemplates(melodyPitchClassCount, templateSize)
    possibleCombinations = combinations(melodyPitchClassCount, templateSize)
    templates = matrix(0, nrow=melodyPitchClassCount, ncol=templateCount)
    templates[cbind(as.vector(t(possibleCombinations)), rep(1:templateCount, each=templateSize))] =     1
    dictionaryMatrix = cbind(dictionaryMatrix, templates)
  }
  
  if(stateTyingMethod %in% c("none", "ignore"))
  {
    cat("Creating a shift-translator for the dictionary...\n")
    templateTranslator = matrix(1, 12, ncol(dictionaryMatrix)) # this will hold information what        template will we get if we shift a given template by given shift; the first column should be all        1s, because we always get the same when we shift nothing
    templateNr = 1
    for(templateSize in 1:maxConcurrentNotes)
    {
      templateCount = howManyPitchTemplates(12, templateSize)
      currentColumns = templateNr + 1:templateCount
      for(colNr in currentColumns)
      {
        for(shift in 0:11)
        {
          # shift the vector right(!) by given amount
          if(shift == 0) {
            shiftedVector = dictionaryMatrix[,colNr]
          } else
            shiftedVector = dictionaryMatrix[c((12-shift+1):12, 1:(12-shift)),colNr]
            for(comparedColNr in currentColumns)
              if(all(dictionaryMatrix[,comparedColNr] == shiftedVector))
              {
                templateTranslator[shift+1, colNr] = comparedColNr
                break
              }
        }
      }
      templateNr = templateNr + templateCount
    }
  }

  save(dictionaryMatrix, templateTranslator, file=outputFilename)
}

preloadDataset = function(dataset, datasetname, unitType, unitDivision)
{
  outputFilename = paste("data/dataset_", datasetname, "_", stateTyingMethod, ".Rdata", sep="")

  if(file.exists(outputFilename))
  {
    cat("Dataset file already exists. To re-calculate, delete file", outputFilename, "\n")
#     return()
  }

  #############################################################################
  ## Loading melody data
  {
    cat("Loading melody data...\n")
    allMelody = list()
    for(filename in dataset)
    {
      cat(getFileNumber(filename), " ", sep="")
      currentData = strsplit(readLines(paste("data/datastreams/", filename, ".melody", sep="")), ", ")
      if(ignoreOnsetMarks)
        currentData = lapply(currentData, function(noteNames) sub("[0-9]+", "", noteNames))
      allMelody = c(allMelody, currentData)
    }
    cat("\n")

    cat("Creating the activity matrix...\n")
    activityMatrix = sapply(allMelody, function(x) 1*(pitchClasses[-melodyPitchClassCount] %in% x))
    if(stateTyingMethod == "ignore")
    {
      activityMatrix = t(sapply(0:11, function(pitchClassNr) colSums(activityMatrix[pitchClassNrs[-melodyPitchClassCount] == pitchClassNr,], na.rm=TRUE)))
      activityMatrix[activityMatrix > 1] = 1
    } else
      rownames(activityMatrix) = pitchClasses[-length(pitchClasses)]

    cat("Finding activity vectors in the dictionary")
    maxConcurrentNotes = 12 # max(colSums(activityMatrix))
    load(paste("data/dictionaryMatrix_", stateTyingMethod, "_", maxConcurrentNotes, ".Rdata", sep=""))
    dictionarySizes = colSums(dictionaryMatrix)
    activitySizes = colSums(activityMatrix)
    templateSequence = rep(NA, ncol(activityMatrix))
    for(vectorNr in 1:ncol(activityMatrix))
    {
      if((vectorNr %% 10000) == 0) cat(".")
      for(templateNr in which(dictionarySizes == activitySizes[vectorNr]))
        if(all(dictionaryMatrix[,templateNr] == activityMatrix[,vectorNr]))
        {
          templateSequence[vectorNr] = templateNr
          break
        }
    }
    cat("\n")
  }
  #############################################################################
  ## Loading chord data
  {
    cat("Loading chord data...\n")
    allChordData = data.frame()
    allChordPrevData = data.frame()
    for(filename in dataset)
    {
      cat(getFileNumber(filename), " ", sep="")
      currentData = read.table(paste("data/datastreams/", filename, ".chord", sep=""), sep=",", comment.char="", stringsAsFactors=FALSE)
      allChordData = rbind(allChordData, cbind(currentData, filename))
      allChordPrevData = rbind(allChordPrevData, c(NA, NA), currentData[-nrow(currentData),])
    }
    cat("\n")
    colnames(allChordData) = c("Root", "Type", "Filename")
    colnames(allChordPrevData) = c("Root", "Type")

    # Data bugfixes
    allChordData[which(allChordData[,2] == "maj69"),2] = "major"  # maj69 is a major triad with major 6th and major 9th added; it is not a valid MusicXML chord
    allChordPrevData[which(allChordPrevData[,2] == "maj69"),2] = "major"

    allChordData[which((!is.na(allChordData[,1])) & (is.na(allChordData[,2]))), 1] = NA
    allChordPrevData[which((!is.na(allChordPrevData[,1])) & (is.na(allChordPrevData[,2]))), 1] = NA
    allChordData[which((!is.na(allChordData[,1])) & (allChordData[,2] == "none")), 1] = NA
    allChordPrevData[which((!is.na(allChordPrevData[,1])) & (allChordPrevData[,2] == "none")), 1] = NA

    allTypes = factor(allChordData[,2], levels=allPossibleChordTypes)
    allTypesPrev = factor(allChordPrevData[,2], levels=allPossibleChordTypes)
    allTypes[is.na(allTypes)] = "none"
    allTypesPrev[is.na(allTypesPrev)] = "none"

    if(any(is.na(allTypes))) stop("Aaarghh!")

    allRoots = allChordData[,1]
    allRootsPrev = allChordPrevData[,1]
    allRoots[is.na(allRoots)] = "none"
    allRootsPrev[is.na(allRootsPrev)] = "none"
    allRoots = pitchClassNrs[allRoots]
    allRootsPrev = pitchClassNrs[allRootsPrev]

    # If we ignore enharmonic pitch spelling, then convert all pitches to a simple, 12-tone scale:
    if(stateTyingMethod == "ignore")
    {
      allRoots[is.na(allRoots)] = 12
      names(allRoots) = possiblePitchClasses[allRoots+1]
      allRootInts = as.integer(allRoots)+1
  
      allRootsPrev[is.na(allRootsPrev)] = 12
      names(allRootsPrev) = possiblePitchClasses[allRootsPrev+1]
      allRootPrevInts = as.integer(allRootsPrev)+1

    } else {
      allRootInts = as.integer(factor(names(allRoots), levels=pitchClasses))
      allRootPrevInts = as.integer(factor(names(allRootsPrev), levels=pitchClasses))
    }
  
    allTypeInts = as.integer(allTypes)
    allTypePrevInts = as.integer(allTypesPrev)
  }
  #############################################################################
  ## Loading tonality data
  {
    cat("Loading tonality data...\n")
    allKeyData = data.frame()
    for(filename in dataset)
    {
      cat(getFileNumber(filename), " ", sep="")
      currentData = read.table(paste("data/datastreams/", filename, ".tonality", sep=""), sep=",", comment.char="", stringsAsFactors=FALSE)
      allKeyData = rbind(allKeyData, data.frame(tonic=factor(paste(c("C","D","E","F","G","A","B")[currentData[,1]+1], c("b","","#")[currentData[,2]+2], sep=""), levels=pitchClasses), mode=currentData[,3]))
    }
    cat("\n")

    if(stateTyingMethod == "ignore")
    {
      # Simplify pitch spelling:
      allKeyData[,1] = factor(possiblePitchClasses[pitchClassNrs[allKeyData[,1]]+1], levels=possiblePitchClasses[1:12])
      possibleKeys = c(paste(possiblePitchClasses[1:12], "major", sep="-"), paste(possiblePitchClasses[1:12], "minor", sep="-"))
    } else
    {
      possibleKeys = c(paste(possibleMajorKeys, "major", sep="-"), paste(possibleMinorKeys, "minor", sep="-"))
    }
  
    allKeys = factor(paste(allKeyData[,1], allKeyData[,2], sep="-"), levels=possibleKeys)
    allKeysInts = as.integer(allKeys)
  }

  save(melodyPitchClassCount, possiblePitchClasses, possibleKeys,
       allRoots, allTypes, allRootInts, allTypeInts,
       activityMatrix, templateSequence,
       allKeysInts,
       allRootsPrev, allTypesPrev, allRootPrevInts, allTypePrevInts,
       file=outputFilename)
}

# Precalculate dictionary matrix
precalculateDictionaryMatrix(5)
precalculateDictionaryMatrix(7)
precalculateDictionaryMatrix(8)
precalculateDictionaryMatrix(12)

# Convert MusicXML to datastreams
if(system(paste("bash", "prepareDataForTraining.sh", unitType, unitDivision)) != 0)
  stop("prepareDataForTraining.sh failed")

# Load datstreams to R
preloadDataset(leadsheets.training, paste("training_", unitType, "_", unitDivision, sep=""), unitType, unitDivision)
preloadDataset(leadsheets.validation, paste("validation_", unitType, "_", unitDivision, sep=""), unitType, unitDivision)
preloadDataset(leadsheets.testing, paste("testing_", unitType, "_", unitDivision, sep=""), unitType, unitDivision)
