# -----------------------------------------------------------------
# 
#   MELODY HARMONIZATION WITH INTERPOLATED PROBABILISTIC MODELS
# 
# This package contains the code for automatic melody harmonization
# described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Melody harmonisation
# with interpolated probabilistic models,” INRIA Technical
# Report RR-8110, 2012,
# URL: http://hal.inria.fr/hal-00742957/PDF/RR-8110.pdf
# 
# This code has been written by:
#  - Stanisław Raczyński, INRIA
#  - Satoru Fukayama, The University of Tokyo
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# Contact information: stanislaw.raczynski@inria.fr
# 
# -----------------------------------------------------------------

source("settings.R")
source("musicology.R")
source("datasetDefinitions.R")

prepareFile = function(filename, unitType=1, unitDivision=1)
{
  system(paste("python", "convertToDataStreams.py", paste(filename, sep="/"),  "tempDataStream", unitType, unitDivision))
  system("sed -i \"s/(//g;s/)//g;s/'//g;s/ //g\" tempDataStream.chord")
  system("sed -i \"s/(//g;s/)//g;s/'//g;s/,$//g\" tempDataStream.melody")
  system("sed -i \"s/(//g;s/)//g;s/'//g;s/ //g\" tempDataStream.tonality")

  # Load the chord sequence (for evaluation)
  chordData = read.table("tempDataStream.chord", sep=",", comment.char="", stringsAsFactors=FALSE)

  allRoots = chordData[,1]
  allRoots[is.na(allRoots)] = "none"
  allRoots = pitchClassNrs[allRoots]
  allRoots[is.na(allRoots)] = 12
  names(allRoots) = possiblePitchClasses[allRoots+1]
  allRootInts = as.integer(allRoots)+1

  allTypes = factor(chordData[,2], levels=allPossibleChordTypes)
  allTypes[is.na(allTypes)] = "none"
  allTypeInts = as.integer(allTypes)

  currentData = strsplit(readLines("tempDataStream.melody"), ", ")
  if(ignoreOnsetMarks)
    currentData = lapply(currentData, function(noteNames) sub("[0-9]+", "", noteNames))

  # Creating the activity matrix
  activityMatrix = sapply(currentData, function(x) 1*(pitchClasses[-melodyPitchClassCount] %in% x))
  if(stateTyingMethod == "ignore")
  {
    activityMatrix = t(sapply(0:11, function(pitchClassNr) colSums(matrix(activityMatrix[pitchClassNrs[-length(pitchClassNrs)] == pitchClassNr,], ncol=ncol(activityMatrix)), na.rm=TRUE)))
    activityMatrix[activityMatrix > 1] = 1
  } else {
    rownames(activityMatrix) = pitchClasses[-length(pitchClasses)]
  }
  if(nrow(activityMatrix) != 12)
    activityMatrix = t(activityMatrix)   # bugfix for 1-row matrices

  # Finding activity vectors in the dictionary
  maxConcurrentNotes = 12 # max(colSums(activityMatrix))
  load(paste("data/dictionaryMatrix_", stateTyingMethod, "_", maxConcurrentNotes, ".Rdata", sep=""))
  dictionarySizes = colSums(dictionaryMatrix)
  activitySizes = colSums(activityMatrix)
  templateSequence = rep(NA, ncol(activityMatrix))
  for(vectorNr in 1:ncol(activityMatrix))
  {
    if((vectorNr %% 10000) == 0) cat(".")
      for(templateNr in which(dictionarySizes == activitySizes[vectorNr]))
        if(all(dictionaryMatrix[,templateNr] == activityMatrix[,vectorNr]))
        {
          templateSequence[vectorNr] = templateNr
          break
        }
  }

  # Load the tonality sequence
  keyData = read.table("tempDataStream.tonality", sep=",", comment.char="", stringsAsFactors=FALSE)
  keyData = data.frame(tonic=factor(paste(c("C","D","E","F","G","A","B")[keyData[,1]+1], c("b","","#")[keyData[,2]+2], sep=""), levels=pitchClasses), mode=keyData[,3])
  keyData[,1] = factor(possiblePitchClasses[pitchClassNrs[keyData[,1]]+1], levels=possiblePitchClasses[1:12])
  possibleKeys = c(paste(possiblePitchClasses[1:12], "major", sep="-"), paste(possiblePitchClasses[1:12], "minor", sep="-"))
  keyData = factor(paste(keyData[,1], keyData[,2], sep="-"), levels=possibleKeys)
  keys = as.integer(keyData)

  unlink("tempDataStream.*")

  return(list(templateSequence, keys, allRootInts, allTypeInts))
}

