# -----------------------------------------------------------------
# 
#   MELODY HARMONIZATION WITH INTERPOLATED PROBABILISTIC MODELS
# 
# This package contains the code for automatic melody harmonization
# described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Melody harmonisation
# with interpolated probabilistic models,” INRIA Technical
# Report RR-8110, 2012,
# URL: http://hal.inria.fr/hal-00742957/PDF/RR-8110.pdf
# 
# This code has been written by:
#  - Stanisław Raczyński, INRIA
#  - Satoru Fukayama, The University of Tokyo
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# Contact information: stanislaw.raczynski@inria.fr
# 
# -----------------------------------------------------------------

source("settings.R")
source("musicology.R")

outputFilename = paste("data/models_", unitType, "_", unitDivision, "_", stateTyingMethod, ".Rdata", sep="")

if(file.exists(outputFilename))
{
  cat("Model file already exists. To re-train, delete file", outputFilename, "\n")
  load(outputFilename)
  load(paste("data/dictionaryMatrix_", stateTyingMethod, "_12.Rdata", sep=""))
} else {
  library("gtools")

  load(paste("data/dataset_training_", unitType, "_", unitDivision, "_", stateTyingMethod, ".Rdata", sep=""))
  load(paste("data/dictionaryMatrix_", stateTyingMethod, "_12.Rdata", sep=""))

  #########################################################################
  ## Prior chord distribution
  {
    cat("Getting prior distribution of chords...\n")
    chordDistribution = sapply(allPossibleChordTypes, function(chordName) table(factor(names(allRoots[allTypes == chordName]), levels=possiblePitchClasses)))
    chordDistribution = chordDistribution / sum(chordDistribution)  # normalization
  }
  ########################################################################
  ## Melody model
  {
    cat("Calculating the chord|melody probability array")
    melodyModel = array(0, dim=c(dim(chordDistribution), ncol(dictionaryMatrix)),
                        dimnames=list(possiblePitchClasses, levels(allTypes), NULL))
    if(stateTyingMethod == "none")
    {
      for(ii in 1:length(templateSequence))
      {
        melodyModel[allRootInts[ii], allTypeInts[ii], templateSequence[ii]] = melodyModel[allRootInts[ii], allTypeInts[ii], templateSequence[ii]] + 1
      } # warning: this cannot be vectorized, because there are repetitions in index triplets
    } else if(stateTyingMethod == "ignore")
    {
      for(ii in 1:length(templateSequence))
      {
        if((ii %% 10000) == 0) cat(".")
        if(allTypes[ii] == "none")  # it's a special case
        {
          templateNrs = templateTranslator[, templateSequence[ii]]
          melodyModel["none", "none", templateNrs] = melodyModel["none", "none", templateNrs] + 1
        } else
        {
          shifts = (0:11 - allRoots[ii]) %% 12 + 1
          for(rootNr in seq_along(possiblePitchClasses))
          {
            templateNr = templateTranslator[shifts[rootNr], templateSequence[ii]]
            melodyModel[rootNr, allTypeInts[ii], templateNr] = melodyModel[rootNr, allTypeInts[ii], templateNr] + 1
          }
        }
      }
    } else if(stateTyingMethod == "usekey")
    {
      stop("Unimplemented.")
    } else
    {
      stop("Unknown state tying method.")
    }
    cat("\n")

    cat("Normalizing the melody model...\n")
    normalizingMatrix = apply(melodyModel, c(3), sum)
    for(ii in 1:dim(melodyModel)[3])
      melodyModel[,,ii] = melodyModel[,,ii] / (normalizingMatrix[ii] + 1e-16)
  }
  ########################################################################
  ## Tonality model
  {
    cat("Calculating the chord|tonality probability array")
    tonalityModel = array(0, dim=c(dim(chordDistribution), length(possibleKeys)),
                            dimnames = list(possiblePitchClasses, levels(allTypes), possibleKeys))
    if(stateTyingMethod == "none")
    {
      cat("...\n")
      for(ii in 1:length(allKeysInts))
      {
        tonalityModel[allRootInts[ii],allTypeInts[ii],allKeysInts[ii]] = tonalityModel[allRootInts[ii],allTypeInts[ii],allKeysInts[ii]] + 1
      }
    } else if(stateTyingMethod == "ignore")
    {
      for(ii in 1:length(allKeysInts))
      {
        if((ii %% 10000) == 0) cat(".")
        if(allTypes[ii] == "none")  # it's a special case
        {
          tonalityModel["none", "none",] = tonalityModel["none", "none",] + 1
        } else {
          for(keyNr in 0:11)
          {
            rootNr = (allRootInts[ii]-allKeysInts[ii]+keyNr)%%12 + 1
            if(allKeysInts[ii] < 13)
              tonalityModel[rootNr,allTypeInts[ii],keyNr+1] = tonalityModel[rootNr,allTypeInts[ii],keyNr+1] + 1
            else
              tonalityModel[rootNr,allTypeInts[ii],keyNr+13] = tonalityModel[rootNr,allTypeInts[ii],keyNr+13] + 1
          }
        }
      }
      cat("\n")
    }

    #TODO: Write a generalized normalization function
    cat("Normalizing the tonality model...\n")
    normalizingMatrix = apply(tonalityModel, c(3), sum)
    for(ii in 1:dim(tonalityModel)[3])
      tonalityModel[,,ii] = tonalityModel[,,ii] / (normalizingMatrix[ii] + 1e-16)
  }
  ########################################################################
  ## Chord bigram model
  {
    cat("Calculating the chord|chord probability array")
    chordModel = array(0, dim=c(dim(chordDistribution), dim(chordDistribution)),
                      dimnames = list(possiblePitchClasses, levels(allTypes),
                                      possiblePitchClasses, levels(allTypes)))
    if(stateTyingMethod == "none")
    {
      cat("...\n")
      for(ii in 1:length(allRoots))
        chordModel[allRootInts[ii], allTypeInts[ii], allRootPrevInts[ii], allTypePrevInts[ii]] = chordModel[allRootInts[ii], allTypeInts[ii], allRootPrevInts[ii], allTypePrevInts[ii]] + 1
    } else if(stateTyingMethod == "ignore")
    {
      for(ii in 1:length(allRoots))
      {
        if((ii %% 10000) == 0) cat(".")
        if((allTypes[ii] == "none") && (allTypesPrev[ii] == "none"))
        {
          chordModel["none","none","none","none"] = chordModel["none","none","none","none"] + 1
        } else if(allTypes[ii] == "none")
        {
          chordModel["none", "none",,allTypePrevInts[ii]] = chordModel["none", "none",,allTypePrevInts[ii]] + 1
        } else if(allTypesPrev[ii] == "none")
        {
          chordModel[,allTypeInts[ii],"none","none"] = chordModel[,allTypeInts[ii],"none","none"] + 1
        } else {
          for(prevRootNr in 0:11)
          {
            rootNr = (allRootInts[ii]-allRootPrevInts[ii]+prevRootNr)%%12 + 1
            chordModel[rootNr, allTypeInts[ii], prevRootNr+1, allTypePrevInts[ii]] = chordModel[rootNr, allTypeInts[ii], prevRootNr+1, allTypePrevInts[ii]] + 1
          }
        }
      }
    }

    cat("Normalizing the chord model...\n")
    normalizingMatrix = apply(chordModel, c(3,4), sum)
    for(ii in 1:dim(chordModel)[3]) for(jj in 1:dim(chordModel)[4])
      chordModel[,,ii,jj] = chordModel[,,ii,jj] / (normalizingMatrix[ii,jj] + 1e-16)
  }
  ########################################################################
  ## Saving the results
  {
    save(melodyModel, tonalityModel, chordModel, chordDistribution, stateTyingMethod, file=outputFilename)
  }
}

########################################################################
## Plotting the results
if(makePlots) {
  cat("Making nice plots...\n")
  chordSortOrder = names(rev(sort(colSums(chordDistribution))))  # chord types sorted according to how frequent they are

  plotChordDistribution = function(chordDistribution, main, filename)
  {
    cat("· Plotting to", filename, "\n")
    pdf(paste("Plots/", filename, ".pdf", sep=""), width=8, height=6)
      par(cex=0.6, mar=c(10,4,4,2)+0.1)
      par(fig=c(0.03,0.94,0.02,0.93))
        image(x=seq_along(chordSortOrder), y=seq_along(possiblePitchClasses), z=log(t(chordDistribution[,chordSortOrder])), col=gray.colors(256,1,0,1), ylab="Root pitch class", xlab="", axes=F, main="", zlim=c(-12,0))
        axis(1, at=seq_along(chordSortOrder), chordSortOrder, las=2)
        axis(2, at=seq_along(possiblePitchClasses), possiblePitchClasses, las=2)
        box()
      chordTypeDist = colSums(chordDistribution)[chordSortOrder]; names(chordTypeDist) = NULL
      chordRootDist = rowSums(chordDistribution); names(chordRootDist) = NULL
      par(fig=c(0,0.97,0.65,1), new=TRUE)
        barplot(chordTypeDist, axes=F)
      par(fig=c(0.85,1,0,0.95), new=TRUE)
        barplot(chordRootDist, axes=F, horiz=T)
    dev.off()
  }

  plotScale = function()
  {
    pdf("Plots/chordScale.pdf", width=0.8, height=6)
      par(cex=0.6, mar=c(11,2,7,2)+0.1)
      image(x=1, y=seq(0,1,length.out=500), z=log(matrix(seq(0,1,length.out=500), nrow=1)), axes=F, xlab="", ylab="", main="", col=gray.colors(256,1,0,1), xlim=c(-12,0))
      axis(4, at=seq(0,1,length.out=11), labels=seq(0,1,length.out=11), las=1)
      box()
    dev.off()
  }

  findInDictionary = function(pitchNames)
  {
    pitchTemplate = table(c(notexisting=1:(length(possiblePitchClasses)-1),sapply(pitchNames, function(pname) which(possiblePitchClasses == pname))))-1
    return(which(sapply(1:ncol(dictionaryMatrix), function(ii) all(dictionaryMatrix[,ii] == pitchTemplate))))
  }

  plotChordPrior = function()
    plotChordDistribution(chordDistribution, "Prior chord log distribution", "priorChordDist")

  plotMelodyModel = function(pitches)
    plotChordDistribution(melodyModel[,chordSortOrder,findInDictionary(pitches)], paste("Chord log distribution if the melody pitches are (", paste(pitches, collapse=","), ")", sep=""), paste("melodyModel_chordDist_", paste(pitches, collapse=""), sep=""))

  plotTonalityModel = function(tonality)
    plotChordDistribution(tonalityModel[,chordSortOrder,tonality], paste("Chord log distribution if the tonality is", tonality), paste("tonalityModel_", tonality, sep=""))

  plotChordModel = function(root, kind)
    plotChordDistribution(chordModel[,chordSortOrder,root,kind], paste("Chord log distribution if the previous chord is ", root, "-", kind, sep=""), paste("chordModel_", root, kind, sep=""))

  plotChordPrior()

  plotMelodyModel("C")
  plotMelodyModel("C#")          # to check state tying
  plotMelodyModel(c("C","E"))
  plotMelodyModel(c("C","E","G"))

  plotTonalityModel("C-major")
  plotTonalityModel("C#-major")  # to check state tying

  plotChordModel("C", "major")   # tonic of C-maj
  plotChordModel("C#", "major")  # to check state tying
  plotChordModel("G", "major")   # dominant of C-maj

  plotScale()
}

