# -----------------------------------------------------------------
# 
#   MELODY HARMONIZATION WITH INTERPOLATED PROBABILISTIC MODELS
# 
# This package contains the code for automatic melody harmonization
# described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Melody harmonisation
# with interpolated probabilistic models,” INRIA Technical
# Report RR-8110, 2012,
# URL: http://hal.inria.fr/hal-00742957/PDF/RR-8110.pdf
# 
# This code has been written by:
#  - Stanisław Raczyński, INRIA
#  - Satoru Fukayama, The University of Tokyo
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# Contact information: stanislaw.raczynski@inria.fr
# 
# -----------------------------------------------------------------

outputFilename = paste("data/fudgeFactors_", unitType, "_", unitDivision, "_", stateTyingMethod, ".Rdata", sep="")

if(file.exists(outputFilename))
{
  cat("Fudge factor file already exists. To re-train, delete file", outputFilename, "\n")
} else {
  source("settings.R")
  source("calculateCrossEntropy.R")

  cat("Training", outputFilename, "\n")

  # 'L-BFGS-B is' faster and better, but hates infinities
  # 'Nelder-Mead' is the slowest, but robust
  optimizeCoefficients = function(interpolation="loglinear",
                                  tonalityModel=TRUE,
                                  melodyModel=TRUE,
                                  bigramModel=TRUE,
                                  coefficientTrainingMethod="L-BFGS-B",
                                  initialValues=NA
                                 )
  {
    placeParameters = createPlacementFunction(interpolation,
                                              tonalityModel=tonalityModel,
                                              melodyModel=melodyModel,
                                              bigramModel=bigramModel)

    modelsUsed = c(tonalityModel, melodyModel, bigramModel)
    if(interpolation == "loglinear") {
      parameterCount = sum(rep(modelsUsed*1, each=3))
    } else if(interpolation == "linear")
      parameterCount = sum(c(modelsUsed*1, 1, 1))

    # Preload the data so that it is not loaded for each interpolation coefficient tuple (waaay faster)
    allData = loadData(dataset="validation", datasetSuffix=paste("_", unitType, "_", unitDivision, "_", stateTyingMethod, sep=""))

    if(is.na(initialValues))
      initialValues = rep(0.5,parameterCount)
    
    # Perform the actual optimization
    # The lower bound for all parameters is 0
    return(optim(par=initialValues,
                 fn=function(x) calculateCrossEntropies(placeParameters(x),
                                                        interpolation=interpolation,
                                                        allData=allData,
                                                        avoidInfinity=TRUE,
                                                        silent=FALSE),
                 method=coefficientTrainingMethod,
                 lower=rep(0,parameterCount),
                 control=list(fnscale=1e-6)    # to avoid the numerical gradient becoming positive in the steepest descent direction and the "ERROR: ABNORMAL_TERMINATION_IN_LNSRCH" error message
                 ))
  }

  # Optimize coefficients for both interpolation methods
  # and model configurations: MTB, M, T, B, MT

  modelCombinations = cbind(
    rep(c(1,1,0,0,1) == 1, 2),
    rep(c(1,0,1,0,1) == 1, 2),
    rep(c(1,0,0,1,0) == 1, 2)
  )
  colnames(modelCombinations) = c("M", "T", "B")
  interpolationMethods = rep(c("linear", "loglinear"),each=nrow(modelCombinations)/2)
  rownames(modelCombinations) = interpolationMethods

  fudgeFactors = mapply(optimizeCoefficients, interpolation=interpolationMethods, melodyModel=modelCombinations[,1], tonalityModel=modelCombinations[,2], bigramModel=modelCombinations[,3])

  # Save results
  save(fudgeFactors, modelCombinations, file=outputFilename)
}
