# -----------------------------------------------------------------
# 
#   MELODY HARMONIZATION WITH INTERPOLATED PROBABILISTIC MODELS
# 
# This package contains the code for automatic melody harmonization
# described in the paper:
# 
# S.A. Raczyński, S. Fukayama, E. Vincent, “Melody harmonisation
# with interpolated probabilistic models,” INRIA Technical
# Report RR-8110, 2012,
# URL: http://hal.inria.fr/hal-00742957/PDF/RR-8110.pdf
# 
# This code has been written by:
#  - Stanisław Raczyński, INRIA
#  - Satoru Fukayama, The University of Tokyo
# 
# The code is licensed under the GNU GENERAL PUBLIC LICENSE
# Version 3, which can be found at:
#   http://www.gnu.org/licenses/gpl-3.0.txt
# 
# Contact information: stanislaw.raczynski@inria.fr
# 
# -----------------------------------------------------------------

library("combinat")

source("harmonizationEvaluation.R")

unitDivision = 1/4
allFiles = loadData(leadsheets.validation, unitDivision=unitDivision)

optimizeHarmonization = function(pars, interpolation, multicore=TRUE, unitDivision=1)
{
  indices = hcube(sapply(pars, length))
  allResults = c(lapply, mclapply)[[multicore+1]](1:nrow(indices), function(ii)
  {
    timesElapsed = system.time({
      singleResult = rowMeans(getHarmonizations(allFiles, allFiles, interpolation, sapply(seq_along(pars), function(jj) pars[[jj]][indices[ii,jj]]), multicore=!multicore, unitDivision=unitDivision)$allResults)[1]
    })[3]
    cat("\nFinished iteration ", ii, "/", nrow(indices), " in ", round(sum(timesElapsed, na.rm=T)/60,1), " minutes: ", singleResult, "\n", sep="")
    return(singleResult)
  })
  bestIndices = indices[which.max(allResults),]
  return(list(pars=sapply(seq_along(pars), function(ii) pars[[ii]][bestIndices[ii]]), results=allResults))
}

# For example
optimizedPars_M_LL = optimizeHarmonization(list(seq(0.05,0.3, length.out=6),
                                                c(0,0.1,0.5),
                                                c(0,0.1,0.5),
                                                0, 0, 1e-16,
                                                0, 0, 1e-16),
                                          "loglinear", unitDivision=unitDivision)
optimizedPars_M_L  = optimizeHarmonization(list(0.5, 0, 0,
                                                c(0,0.1,0.5),
                                                c(0,0.1,0.5)),
                                          "linear", unitDivision=unitDivision)
